/* Copyright 2023 CMU, Facebook, LANL, MIT, NVIDIA, and Stanford (alphabetical)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "flexflow/parallel_ops/replicate.h"
#include "flexflow/model.h"
#include "flexflow/parallel_ops/kernels/replicate_kernels.h"
#include "flexflow/utils/hash_utils.h"

namespace FlexFlow {
// declare Legion names
using Legion::ArgumentMap;
using Legion::Context;
using Legion::coord_t;
using Legion::Domain;
using Legion::FutureMap;
using Legion::IndexLauncher;
using Legion::LogicalPartition;
using Legion::LogicalRegion;
using Legion::Machine;
using Legion::Memory;
using Legion::PhysicalRegion;
using Legion::Predicate;
using Legion::Rect;
using Legion::RegionRequirement;
using Legion::Runtime;
using Legion::Task;
using Legion::TaskArgument;
using Legion::TaskLauncher;

using namespace FlexFlow::Kernels::Replicate;

/* Params */
bool operator==(ReplicateParams const &lhs, ReplicateParams const &rhs) {
  return lhs.replicate_legion_dim == rhs.replicate_legion_dim &&
         lhs.replicate_degree == rhs.replicate_degree &&
         std::strcmp(lhs.name, rhs.name) == 0;
}

bool ReplicateParams::is_valid(ParallelTensorShape const &input) const {
  return input.is_valid();
}

ReplicateParams Replicate::get_params() const {
  ReplicateParams params;
  params.replicate_legion_dim = this->replicate_dim;
  params.replicate_degree = this->replicate_degree;
  if (strlen(this->name) < MAX_OPNAME) {
    strcpy(params.name, this->name);
  }
  return params;
}

Replicate::Replicate(FFModel &model,
                     const ParallelTensor _input,
                     int _replicate_legion_dim,
                     int _replicate_degree,
                     char const *name)
    : ParallelOp(model, OP_REPLICATE, name, _input),
      replicate_dim(_replicate_legion_dim),
      replicate_degree(_replicate_degree) {
  int numdim = _input->num_dims;
  ParallelDim dims[MAX_TENSOR_DIM];
  for (int i = 0; i < numdim; i++) {
    dims[i] = _input->dims[i];
  }
  dims[replicate_dim].size *= replicate_degree;
  dims[replicate_dim].degree *= replicate_degree;
  ParallelTensorBase::update_parallel_ids(numdim, dims);
  outputs[0] = model.create_parallel_tensor_legion_ordering(
      numdim, dims, _input->data_type, this);
  // inputs[0]->print("Replicate::input");
  // outputs[0]->print("Replicate::output");
}

Replicate::Replicate(FFModel &model,
                     ReplicateParams const &params,
                     ParallelTensor const input,
                     char const *name)
    : Replicate(model,
                input,
                params.replicate_legion_dim,
                params.replicate_degree,
                params.name) {}

void Replicate::create_input_partition(FFModel &ff) {
  assert(outputs[0]->part != LogicalPartition::NO_PART);
  assert(inputs[0]->part != LogicalPartition::NO_PART);
  // input_lp is an aliased partitioning along the replica dim
  ff.create_aliased_partition(outputs[0]->num_dims,
                              outputs[0]->dims,
                              replicate_dim,
                              outputs[0]->parallel_is,
                              inputs[0]->region,
                              input_lp);
  // output_grad_lp is a disjoint partition
  ff.create_disjoint_partition(inputs[0]->num_dims,
                               inputs[0]->dims,
                               inputs[0]->parallel_is,
                               outputs[0]->region_grad,
                               output_grad_lp);
}

void Replicate::create_input_partition_inference(
    FFModel &ff,
    std::vector<ParallelTensor> const &batch_inputs,
    std::vector<ParallelTensor> const &batch_outputs) {
  assert(ff.config.computationMode == COMP_MODE_INFERENCE);
  assert(batch_outputs[0]->part != LogicalPartition::NO_PART);
  assert(batch_inputs[0]->part != LogicalPartition::NO_PART);
  // input_lp is an aliased partitioning along the replica dim
  ff.create_aliased_partition(batch_outputs[0]->num_dims,
                              batch_outputs[0]->dims,
                              replicate_dim,
                              batch_outputs[0]->parallel_is,
                              batch_inputs[0]->region,
                              inference_input_lps[batch_inputs[0]]);
  // output_grad_lp is a disjoint partition
  ff.create_disjoint_partition(batch_inputs[0]->num_dims,
                               batch_inputs[0]->dims,
                               batch_inputs[0]->parallel_is,
                               batch_outputs[0]->region_grad,
                               inference_output_grad_lps[batch_outputs[0]]);
}

OpMeta *Replicate::init_task(Task const *task,
                             std::vector<PhysicalRegion> const &regions,
                             Context ctx,
                             Runtime *runtime) {
  Replicate *repl = (Replicate *)task->args;
  FFHandler handle = *((FFHandler const *)task->local_args);
  ReplicateMeta *meta = new ReplicateMeta(handle, repl);
  meta->input_type[0] = repl->inputs[0]->data_type;
  meta->output_type[0] = repl->outputs[0]->data_type;
  assert(meta->input_type[0] == meta->output_type[0]);
  std::strcpy(meta->op_name, repl->name);
  return meta;
}

void Replicate::init_inference(FFModel const &ff,
                               std::vector<ParallelTensor> const &batch_inputs,
                               std::vector<ParallelTensor> const &batch_outputs,
                               MachineView const *mv) {
  parallel_is = batch_outputs[0]->parallel_is;
  ArgumentMap argmap;
  Context ctx = ff.config.lg_ctx;
  Runtime *runtime = ff.config.lg_hlr;
  assert(numOutputs == 1);
  assert(numInputs == 1);
  size_t machine_view_hash =
      mv ? mv->hash() : batch_outputs[0]->machine_view.hash();
  set_argumentmap_for_init_inference(ff, argmap, batch_outputs[0]);
  IndexLauncher launcher(REPLICATE_INIT_TASK_ID,
                         batch_outputs[0]->parallel_is,
                         TaskArgument(this, sizeof(Replicate)),
                         argmap,
                         Predicate::TRUE_PRED,
                         false /*must*/,
                         0 /*mapper_id*/,
                         machine_view_hash);
  launcher.add_region_requirement(
      RegionRequirement(inference_input_lps[batch_inputs[0]],
                        0 /*projection id*/,
                        READ_ONLY,
                        EXCLUSIVE,
                        batch_inputs[0]->region));
  launcher.add_field(0, FID_DATA);
  launcher.add_region_requirement(RegionRequirement(batch_outputs[0]->part,
                                                    0 /*projection id*/,
                                                    WRITE_ONLY,
                                                    EXCLUSIVE,
                                                    batch_outputs[0]->region));
  launcher.add_field(1, FID_DATA);
  FutureMap fm = runtime->execute_index_space(ctx, launcher);
  fm.wait_all_results();
  set_opmeta_from_futuremap_inference(ff, fm, batch_outputs[0]);
}

void Replicate::init(FFModel const &ff) {
  parallel_is = outputs[0]->parallel_is;
  ArgumentMap argmap;
  Context ctx = ff.config.lg_ctx;
  Runtime *runtime = ff.config.lg_hlr;
  assert(numOutputs == 1);
  assert(numInputs == 1);
  set_argumentmap_for_init(ff, argmap);
  IndexLauncher launcher(REPLICATE_INIT_TASK_ID,
                         outputs[0]->parallel_is,
                         TaskArgument(this, sizeof(Replicate)),
                         argmap,
                         Predicate::TRUE_PRED,
                         false /*must*/,
                         0 /*mapper_id*/,
                         outputs[0]->machine_view.hash());
  launcher.add_region_requirement(RegionRequirement(
      input_lp, 0 /*projection id*/, READ_ONLY, EXCLUSIVE, inputs[0]->region));
  launcher.add_field(0, FID_DATA);
  launcher.add_region_requirement(RegionRequirement(outputs[0]->part,
                                                    0 /*projection id*/,
                                                    WRITE_ONLY,
                                                    EXCLUSIVE,
                                                    outputs[0]->region));
  launcher.add_field(1, FID_DATA);
  FutureMap fm = runtime->execute_index_space(ctx, launcher);
  fm.wait_all_results();
  set_opmeta_from_futuremap(ff, fm);
}

FutureMap Replicate::inference(FFModel const &ff,
                               BatchConfigFuture const &bc,
                               std::vector<ParallelTensor> const &batch_inputs,
                               std::vector<ParallelTensor> const &batch_outputs,
                               MachineView const *mv) {
  ArgumentMap argmap;
  Context ctx = ff.config.lg_ctx;
  Runtime *runtime = ff.config.lg_hlr;
  parallel_is = batch_outputs[0]->parallel_is;
  assert(numOutputs == 1);
  assert(numInputs == 1);
  DataType data_type = batch_inputs[0]->data_type;
  size_t machine_view_hash =
      mv ? mv->hash() : batch_outputs[0]->machine_view.hash();
  set_argumentmap_for_inference(ff, argmap, batch_outputs[0]);
  IndexLauncher launcher(REPLICATE_FWD_TASK_ID,
                         batch_outputs[0]->parallel_is,
                         TaskArgument(NULL, 0),
                         argmap,
                         Predicate::TRUE_PRED,
                         false /*must*/,
                         0 /*mapper_id*/,
                         machine_view_hash);
  launcher.add_region_requirement(
      RegionRequirement(inference_input_lps[batch_inputs[0]],
                        0 /*projection id*/,
                        READ_ONLY,
                        EXCLUSIVE,
                        batch_inputs[0]->region));
  launcher.add_field(0, FID_DATA);
  launcher.add_region_requirement(RegionRequirement(batch_outputs[0]->part,
                                                    0 /*projection id*/,
                                                    WRITE_ONLY,
                                                    EXCLUSIVE,
                                                    batch_outputs[0]->region));
  launcher.add_field(1, FID_DATA);
  return runtime->execute_index_space(ctx, launcher);
}

void Replicate::forward(FFModel const &ff) {
  ArgumentMap argmap;
  Context ctx = ff.config.lg_ctx;
  Runtime *runtime = ff.config.lg_hlr;
  parallel_is = outputs[0]->parallel_is;
  assert(numOutputs == 1);
  assert(numInputs == 1);
  set_argumentmap_for_forward(ff, argmap);
  IndexLauncher launcher(REPLICATE_FWD_TASK_ID,
                         outputs[0]->parallel_is,
                         TaskArgument(NULL, 0),
                         argmap,
                         Predicate::TRUE_PRED,
                         false /*must*/,
                         0 /*mapper_id*/,
                         outputs[0]->machine_view.hash());
  launcher.add_region_requirement(RegionRequirement(
      input_lp, 0 /*projection id*/, READ_ONLY, EXCLUSIVE, inputs[0]->region));
  launcher.add_field(0, FID_DATA);
  launcher.add_region_requirement(RegionRequirement(outputs[0]->part,
                                                    0 /*projection id*/,
                                                    WRITE_ONLY,
                                                    EXCLUSIVE,
                                                    outputs[0]->region));
  launcher.add_field(1, FID_DATA);
  runtime->execute_index_space(ctx, launcher);
}

FutureMap Replicate::peft_bwd(FFModel const &ff,
                              BatchConfigFuture const &bc,
                              std::vector<ParallelTensor> const &batch_inputs,
                              std::vector<ParallelTensor> const &batch_outputs,
                              MachineView const *mv) {
  ArgumentMap argmap;
  Context ctx = ff.config.lg_ctx;
  Runtime *runtime = ff.config.lg_hlr;
  assert(numOutputs == 1);
  assert(numInputs == 1);
  assert(batch_inputs[0]->data_type == batch_outputs[0]->data_type);
  DataType data_type = batch_inputs[0]->data_type;

  // Warning: we need to use batch_inputs[0] here, instead of the usual
  // batch_outputs[0]
  parallel_is = batch_inputs[0]->parallel_is;
  MachineView const *view = mv ? mv : &batch_inputs[0]->machine_view;

  set_argumentmap_for_inference(ff, argmap, batch_outputs[0]);
  size_t machine_view_hash = view->hash();
  IndexLauncher launcher(REPLICATE_PEFT_BWD_TASK_ID,
                         parallel_is,
                         TaskArgument(NULL, 0),
                         argmap,
                         Predicate::TRUE_PRED,
                         false /*must*/,
                         0 /*mapper_id*/,
                         machine_view_hash);
  launcher.add_region_requirement(
      RegionRequirement(inference_output_grad_lps[batch_outputs[0]],
                        0 /*projection id*/,
                        READ_ONLY,
                        EXCLUSIVE,
                        batch_outputs[0]->region_grad));
  launcher.add_field(0, FID_DATA);
  launcher.add_region_requirement(
      RegionRequirement(batch_inputs[0]->part_grad,
                        0 /*projection id*/,
                        READ_WRITE,
                        EXCLUSIVE,
                        batch_inputs[0]->region_grad));
  launcher.add_field(1, FID_DATA);
  return runtime->execute_index_space(ctx, launcher);
}

void Replicate::backward(FFModel const &ff) {
  ArgumentMap argmap;
  Context ctx = ff.config.lg_ctx;
  Runtime *runtime = ff.config.lg_hlr;
  assert(numOutputs == 1);
  assert(numInputs == 1);
  IndexLauncher launcher(REPLICATE_BWD_TASK_ID,
                         inputs[0]->parallel_is,
                         TaskArgument(NULL, 0),
                         argmap,
                         Predicate::TRUE_PRED,
                         false /*must*/,
                         0 /*mapper_id*/,
                         inputs[0]->machine_view.hash());
  launcher.add_region_requirement(RegionRequirement(output_grad_lp,
                                                    0 /*projection id*/,
                                                    READ_ONLY,
                                                    EXCLUSIVE,
                                                    outputs[0]->region_grad));
  launcher.add_field(0, FID_DATA);
  launcher.add_region_requirement(RegionRequirement(inputs[0]->part_grad,
                                                    0 /*projection id*/,
                                                    READ_WRITE,
                                                    EXCLUSIVE,
                                                    inputs[0]->region_grad));
  launcher.add_field(1, FID_DATA);
  runtime->execute_index_space(ctx, launcher);
}

bool Replicate::measure_operator_cost(Simulator *sim,
                                      MachineView const &pc,
                                      CostMetrics &cost_metrics) const {
  cost_metrics = CostMetrics();
  cost_metrics.forward_time = 0.0f;
  cost_metrics.backward_time = 0.0f;

  cost_metrics.sync_time = 0;
  cost_metrics.inputs_memory = 0;
  cost_metrics.outputs_memory = 0;
  cost_metrics.weights_memory = 0;
  return true;
}

bool Replicate::get_int_parameter(PMParameter para, int *value) const {
  switch (para) {
    case PM_REPLICATE_DIM:
      *value = replicate_dim;
      return true;
    case PM_REPLICATE_DEGREE:
      *value = replicate_degree;
      return true;
    default:
      return Op::get_int_parameter(para, value);
  }
}

bool Replicate::append_parallel_op_info(
    std::vector<ParallelOpInfo> &parallel_ops) const {
  ParallelOpInfo ret;
  ret.op_type = op_type;
  ret.parallel_dim = replicate_dim;
  ret.parallel_degree = replicate_degree;
  parallel_ops.push_back(ret);
  return true;
}

void Replicate::forward_task(Task const *task,
                             std::vector<PhysicalRegion> const &regions,
                             Context ctx,
                             Runtime *runtime) {
  assert(regions.size() == 2);
  assert(task->regions.size() == 2);

  ReplicateMeta const *m = *((ReplicateMeta **)task->local_args);
  if (m->inference_debugging) {
    std::cout << "INF " << m->op_name << std::endl;
  }

  Domain input_domain = runtime->get_index_space_domain(
      ctx, task->regions[0].region.get_index_space());
  Domain output_domain = runtime->get_index_space_domain(
      ctx, task->regions[1].region.get_index_space());
  // Currently only support the outter most dimension
  for (int i = 0; i < output_domain.get_dim() - 1; i++) {
    assert(output_domain.lo()[i] == input_domain.lo()[i]);
    assert(output_domain.hi()[i] == input_domain.hi()[i]);
  }
  assert(input_domain.get_volume() == output_domain.get_volume());

  GenericTensorAccessorR input = helperGetGenericTensorAccessorRO(
      m->input_type[0], regions[0], task->regions[0], FID_DATA, ctx, runtime);
  GenericTensorAccessorW output = helperGetGenericTensorAccessorWO(
      m->output_type[0], regions[1], task->regions[1], FID_DATA, ctx, runtime);

  assert(input.data_type == output.data_type);

  if (input.data_type == DT_HALF) {
    forward_kernel<half>(
        input.get_half_ptr(), output.get_half_ptr(), input_domain.get_volume());
  } else if (input.data_type == DT_FLOAT) {
    forward_kernel<float>(input.get_float_ptr(),
                          output.get_float_ptr(),
                          input_domain.get_volume());
  } else {
    assert(false && "Unspported data type");
  }
}

void Replicate::peft_bwd_task(Task const *task,
                              std::vector<PhysicalRegion> const &regions,
                              Context ctx,
                              Runtime *runtime) {
  assert(regions.size() == 2);
  assert(task->regions.size() == 2);
  Domain output_grad_domain = runtime->get_index_space_domain(
      ctx, task->regions[0].region.get_index_space());
  Domain input_grad_domain = runtime->get_index_space_domain(
      ctx, task->regions[1].region.get_index_space());
  // Currently only support the outter most dimension
  for (int i = 0; i < output_grad_domain.get_dim() - 1; i++) {
    assert(output_grad_domain.lo()[i] == input_grad_domain.lo()[i]);
    assert(output_grad_domain.hi()[i] == input_grad_domain.hi()[i]);
  }
  size_t num_elements = input_grad_domain.get_volume();
  size_t num_replicas = output_grad_domain.get_volume() / num_elements;
  float const *output_grad_ptr = helperGetTensorPointerRO<float>(
      regions[0], task->regions[0], FID_DATA, ctx, runtime);
  float *input_grad_ptr = helperGetTensorPointerRW<float>(
      regions[1], task->regions[1], FID_DATA, ctx, runtime);

  ReplicateMeta const *m = *((ReplicateMeta **)task->local_args);
  if (m->inference_debugging) {
    std::cout << "BWD " << m->op_name << std::endl;
  }

  backward_kernel<float>(
      output_grad_ptr, input_grad_ptr, num_elements, num_replicas);
}

void Replicate::backward_task(Task const *task,
                              std::vector<PhysicalRegion> const &regions,
                              Context ctx,
                              Runtime *runtime) {
  assert(regions.size() == 2);
  assert(task->regions.size() == 2);
  Domain output_grad_domain = runtime->get_index_space_domain(
      ctx, task->regions[0].region.get_index_space());
  Domain input_grad_domain = runtime->get_index_space_domain(
      ctx, task->regions[1].region.get_index_space());
  // Currently only support the outter most dimension
  for (int i = 0; i < output_grad_domain.get_dim() - 1; i++) {
    assert(output_grad_domain.lo()[i] == input_grad_domain.lo()[i]);
    assert(output_grad_domain.hi()[i] == input_grad_domain.hi()[i]);
  }
  size_t num_elements = input_grad_domain.get_volume();
  size_t num_replicas = output_grad_domain.get_volume() / num_elements;
  float const *output_grad_ptr = helperGetTensorPointerRO<float>(
      regions[0], task->regions[0], FID_DATA, ctx, runtime);
  float *input_grad_ptr = helperGetTensorPointerRW<float>(
      regions[1], task->regions[1], FID_DATA, ctx, runtime);

  backward_kernel<float>(
      output_grad_ptr, input_grad_ptr, num_elements, num_replicas);
}

}; // namespace FlexFlow

namespace std {
size_t hash<FlexFlow::ReplicateParams>::operator()(
    FlexFlow::ReplicateParams const &params) const {
  size_t key = 0;
  hash_combine(key, params.replicate_legion_dim);
  hash_combine(key, params.replicate_degree);
  return key;
}
}; // namespace std
